"""
Volumes API calls.
"""

def getVolumes(self, organizationId=None, volumeId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getVolumes",
            "variables": {
                "organizationId": organizationId,
                "volumeId": volumeId
            },
            "query": """query 
                getVolumes($organizationId: String, $volumeId: String) {
                    getVolumes(organizationId: $organizationId, volumeId: $volumeId) {
                        volumeId
                        organizationId
                        organizationName
                        name
                        updatedAt
                    }
                }"""})
    return self.errorhandler(response, "getVolumes")


def getManagedVolumes(self, organizationId, volumeId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getManagedVolumes",
            "variables": {
                "organizationId": organizationId,
                "volumeId": volumeId
            },
            "query": """query 
                getManagedVolumes($organizationId: String, $volumeId: String) {
                    getManagedVolumes(organizationId: $organizationId, volumeId: $volumeId) {
                        volumeId
                        organizationId
                        name
                        createdAt
                        updatedAt
                        organizations {
                            organizationId
                            name
                        }
                    }
                }"""})
    return self.errorhandler(response, "getManagedVolumes")


def getVolumeData(self, volumeId, dir=None, keys=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getVolumeData",
            "variables": {
                "volumeId": volumeId,
                "keys": keys,
                "dir": dir
            },
            "query": """query 
                getVolumeData($volumeId: String!, $keys: [String]!) {
                    getVolumeData(volumeId: $volumeId, keys: $keys, dir: $dir){
                        key
                        size
                        updatedAt
                        hash
                        url
                    }
                }"""})
    return self.errorhandler(response, "getVolumeData")


def createManagedVolume(self, organizationId, name):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createManagedVolume",
            "variables": {
                "organizationId": organizationId,
                "name": name
            },
            "query": """mutation 
                createManagedVolume($organizationId: String!, $name: String!) {
                    createManagedVolume(organizationId: $organizationId, name: $name)
                }"""})
    return self.errorhandler(response, "createManagedVolume")


def deleteManagedVolume(self, volumeId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteManagedVolume",
            "variables": {
                "volumeId": volumeId,
            },
            "query": """mutation 
                deleteManagedVolume($volumeId: String!) {
                    deleteManagedVolume(volumeId: $volumeId)
                }"""})
    return self.errorhandler(response, "deleteManagedVolume")


def editManagedVolume(self, volumeId, name=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editManagedVolume",
            "variables": {
                "volumeId": volumeId,
                "name": name
            },
            "query": """mutation 
                editManagedVolume($volumeId: String!, $name: String!) {
                    editManagedVolume(volumeId: $volumeId, name: $name)
                }"""})
    return self.errorhandler(response, "editManagedVolume")


def addVolumeOrganization(self, volumeId, organizationId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "addVolumeOrganization",
            "variables": {
                "volumeId": volumeId,
                "organizationId": organizationId
            },
            "query": """mutation 
                addVolumeOrganization($volumeId: String!, $organizationId: String!) {
                    addVolumeOrganization(volumeId: $volumeId, organizationId: $organizationId)
                }"""})
    return self.errorhandler(response, "addVolumeOrganization")


def removeVolumeOrganization(self, volumeId, organizationId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "removeVolumeOrganization",
            "variables": {
                "volumeId": volumeId,
                "organizationId": organizationId
            },
            "query": """mutation 
                removeVolumeOrganization($volumeId: String!, $organizationId: String!) {
                    removeVolumeOrganization(volumeId: $volumeId, organizationId: $organizationId)
                }"""})
    return self.errorhandler(response, "removeVolumeOrganization")


def putVolumeData(self, volumeId, keys=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "putVolumeData",
            "variables": {
                "volumeId": volumeId,
                "keys": keys
            },
            "query": """mutation 
                putVolumeData($volumeId: String!, $keys: [String]!) {
                    putVolumeData(volumeId: $volumeId, keys: $keys){
                        key
                        url
                        fields {
                            key
                            bucket
                            algorithm
                            credential
                            date
                            token
                            policy
                            signature
                        }
                    }
                }"""})
    return self.errorhandler(response, "putVolumeData")


def deleteVolumeData(self, volumeId, keys=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteVolumeData",
            "variables": {
                "volumeId": volumeId,
                "keys": keys
            },
            "query": """mutation 
                deleteVolumeData($volumeId: String!, $keys: [String]!) {
                    deleteVolumeData(volumeId: $volumeId, keys: $keys)
                }"""})
    return self.errorhandler(response, "deleteVolumeData")


def mountVolumes(self, volumes):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "mountVolumes",
            "variables": {
                "volumes": volumes
            },
            "query": """mutation 
                mountVolumes($volumes: [String]!) {
                    mountVolumes(volumes: $volumes){
                        keys
                        rw
                        credentials {
                            accesskeyid
                            accesskey
                            sessiontoken
                        }
                    }
                }"""})
    return self.errorhandler(response, "mountVolumes")


"""
Graphs API calls.
"""

def getGraphs(self, workspaceId, graphId=None, name=None, email=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getGraphs",
            "variables": {
                "workspaceId": workspaceId,
                "graphId": graphId,
                "name": name,
                "email": email
            },
            "query": """query 
                getGraphs($workspaceId: String!, $graphId: String, $name: String, $email: String) {
                    getGraphs(workspaceId: $workspaceId, graphId: $graphId, name: $name, member: $email) {
                        graphId:graphid
                        name
                        channelId
                        user
                        description
                    }
                }"""})
    return self.errorhandler(response, "getGraphs")


def createGraph(self, workspaceId, channelId, graph, name, description):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createGraph",
            "variables": {
                "workspaceId": workspaceId,
                "channelId": channelId,
                "graph": graph,
                "name": name,
                "description": description
            },
            "query": """mutation 
                createGraph($workspaceId: String!, $channelId: String!, $graph: String!, $name: String!, $description: String) {
                    createGraph(workspaceId: $workspaceId, channelId: $channelId, graph: $graph, name: $name, description: $description)
                }"""})
    return self.errorhandler(response, "createGraph")


def deleteGraph(self, workspaceId, graphId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteGraph",
            "variables": {
                "workspaceId": workspaceId,
                "graphId": graphId
            },
            "query": """mutation 
                deleteGraph($workspaceId: String!, $graphId: String!) {
                    deleteGraph(workspaceId: $workspaceId, graphId: $graphId)
                }"""})
    return self.errorhandler(response, "deleteGraph")


def editGraph(self, workspaceId, graphId, name=None, description=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editGraph",
            "variables": {
                "workspaceId": workspaceId,
                "graphId": graphId,
                "name": name,
                "description": description
            },
            "query": """mutation 
                editGraph($workspaceId: String!, $graphId: String!, $name: String, $description: String) {
                    editGraph(workspaceId: $workspaceId, graphId: $graphId, name: $name, description: $description)
                }"""})
    return self.errorhandler(response, "editGraph")


def downloadGraph(self, workspaceId, graphId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "downloadGraph",
            "variables": {
                "workspaceId": workspaceId,
                "graphId": graphId
            },
            "query": """mutation 
                downloadGraph($workspaceId: String!, $graphId: String!) {
                    downloadGraph(workspaceId: $workspaceId, graphId: $graphId)
                }"""})
    return self.errorhandler(response, "downloadGraph")


def getDefaultGraph(self, channelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDefaultGraph",
            "variables": {
                "channelId": channelId,
            },
            "query": """query 
                getDefaultGraph($channelId: String!) {
                    getDefaultGraph(channelId: $channelId)
                }"""})
    return self.errorhandler(response, "getDefaultGraph")
    


"""
Annotations API calls.
"""

def getAnnotations(self, workspaceId, datasetId, annotationId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getAnnotations",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "annotationId": annotationId,
            },
            "query": """query 
                getAnnotations($workspaceId: String!, $datasetId: String! $annotationId: String!) {
                    getAnnotations(workspaceId: $workspaceId, datasetId: $datasetId, annotationId: $annotationId){
                        workspaceId
                        datasetId
                        annotationId
                        map
                        format
                        status
                    }
                }"""})
    return self.errorhandler(response, "getAnnotations")


def getAnnotationFormats(self):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getAnnotationFormats",
            "variables": {},
            "query": """query 
                getAnnotationFormats{
                    getAnnotationFormats
                }"""})
    return self.errorhandler(response, "getAnnotationFormats")


def getAnnotationMaps(self, channelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getAnnotationMaps",
            "variables": {
                "channelId": channelId,
            },
            "query": """query 
                getAnnotationMaps($channelId: String!) {
                    getAnnotationMaps(channelId: $channelId)
                }"""})
    return self.errorhandler(response, "getAnnotationMaps")


def createAnnotation(self, workspaceId, datasetId, format, map):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createAnnotation",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "format": format,
                "map": map
            },
            "query": """mutation 
                createAnnotation($workspaceId: String!, $datasetId: String!, $format: String!, $map: String!) {
                    createAnnotation(workspaceId: $workspaceId, datasetId: $datasetId, format: $format, map: $map)
                }"""})
    return self.errorhandler(response, "createAnnotation")


def downloadAnnotation(self, workspaceId, datasetId, annotationId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "downloadAnnotation",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "annotationId": annotationId
            },
            "query": """mutation 
                downloadAnnotation($workspaceId: String!, $datasetId: String!, $annotationId: String!) {
                    downloadAnnotation(workspaceId: $workspaceId, datasetId: $datasetId, annotationId: $annotationId)
                }"""})
    return self.errorhandler(response, "downloadAnnotation")


def deleteAnnotation(self, workspaceId, annotationId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteAnnotation",
            "variables": {
                "workspaceId": workspaceId,
                "annotationId": annotationId
            },
            "query": """mutation 
                deleteAnnotation($workspaceId: String!, $annotationId: String!) {
                    deleteAnnotation(workspaceId: $workspaceId, annotationId: $annotationId)
                }"""})
    return self.errorhandler(response, "deleteAnnotation")



"""
Members API calls.
"""

def getMembers(self, organizationId=None, workspaceId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getMembers",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId
            },
            "query": """query 
                getMembers($organizationId: String, $workspaceId: String) {
                    getMembers(organizationId: $organizationId, workspaceId: $workspaceId) {
                        organizationId
                        workspaceId
                        userId
                        email
                        name
                        role
                    }
                }"""})
    return self.errorhandler(response, "getMembers")

def getInvitations(self, organizationId=None, workspaceId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getInvitations",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId
            },
            "query": """query 
                getInvitations($organizationId: String, $workspaceId: String) {
                    getInvitations(organizationId: $organizationId, workspaceId: $workspaceId) {
                        invitationId
                        email
                        timestamp
                        organizationId
                        role
                        workspaceId
                    }
                }"""})
    return self.errorhandler(response, "getInvitations")


def addMember(self, email, role=None, organizationId=None, workspaceId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "addMember",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "email": email,
                "role": role
            },
            "query": """mutation 
                addMember($organizationId: String, $workspaceId: String, $email: String!, $role: String) {
                    addMember(organizationId: $organizationId, workspaceId: $workspaceId, email: $email, role: $role)
                }"""})
    return self.errorhandler(response, "addMember")


def removeMember(self, email, organizationId=None, workspaceId=None, invitationId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "removeMember",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "email": email,
                "invitationId":invitationId
            },
            "query": """mutation 
                removeMember($organizationId: String, $workspaceId: String, $email: String!, $invitationId: String) {
                    removeMember(organizationId: $organizationId, workspaceId: $workspaceId, email: $email, invitationId: $invitationId)
                }"""})
    return self.errorhandler(response, "removeMember")


def editMember(self, email, organizationId, role):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editMember",
            "variables": {
                "organizationId": organizationId,
                "email": email,
                "role": role
            },
            "query": """mutation 
                editMember($organizationId: String!, $email: String!, $role: String!) {
                    editMember(organizationId: $organizationId, email: $email, role: $role)
                }"""})
    return self.errorhandler(response, "editMember")



"""anatools is Rendered.ai’s SDK for connecting to the Platform API. 

- The ``anatools.anaclient`` package exposes the Platform API to the User. Everything that a user can do through the Platform Web Interface can be done via the SDK through the ``anatools.anaclient`` package.
- The ``anatools.annotations`` package allows a User to generate annotations locally without the need for the Platform.
- The ``anatools.lib`` package and ``anatools.nodes`` package are used for Channel developers. More info to come!

The SDK's submodule methods are all exposed to the user, so the full path is not required to call a specific method. 

View the `Introduction to Rendered.ai Documentation`_ to learn more.

.. _Introduction to Rendered.ai Documentation:
        https://support.rendered.ai/gc/Introduction-to-Rendered.ai.1577812005.html

"""

from .anaclient.anaclient import client
from .annotations.annotations import annotations

__version__ = '2.1.8'



from anatools.annotations.convert_sagemaker import convert_sagemaker_od
from .convert_coco import convert_coco
from .convert_kitti import convert_kitti
from .convert_pascal import convert_pascal
from .convert_sagemaker import convert_sagemaker_od, convert_sagemaker_ss
from .draw import draw
import yaml

class annotations:
    """Generates annotations given a dataset directory, an output directory and mapping file.
    The dataset directory must include the Ana annotations, images and metadata folders.
    Examples of mapfiles are in the example channel at /ana/channels/example/mapfiles/.
    """

    def bounding_box_2d(self, image_path, out_dir, object_ids=None, object_types=None, line_thickness=1):
        """
        Generates images annotated with 2d bounding boxes for datasets downloaded from the Platform. Optional filter on 
        object_ids or object_types (must choose a single filter).
        
        Parameters
        ----------
        image_path : str
            Path to of specific image file to draw the boxes for.  
        out_dir : str
            File path to directory where the image should be saved to.
        object_ids : list[int]
            List of object id's to annotate. If not provided, all objects will get annotated. Choose either id or type filter.
        object_types: list[str]
            Filter for the object types to annotate. If not provided, all object types will get annotated. Choose either id or type filter.
        line_thickness: int
            Desired line thickness for box outline. 
        """
        draw(image_path, out_dir, draw_type='box_2d', object_ids=object_ids, object_types=object_types, line_thickness=line_thickness)


    def bounding_box_3d(self, image_path, out_dir, object_ids=None, object_types=None, line_thickness=1):
        """
        Generates images annotated with 3d bounding boxes for datasets downloaded from the Platform. Optional filter on 
        object_ids or object_types (must choose a single filter).
        
        Parameters
        ----------
        image_path : str
            Path to of specific image file to draw the boxes for.  
        out_dir : str
            File path to directory where the image should be saved to.
        object_ids : list[int]
            List of object id's to annotate. If not provided, all objects will get annotated. Choose either id or type filter.
        object_types: list[str]
            Filter for the object types to annotate. If not provided, all object types will get annotated. Choose either id or type filter.
        line_thickness: int
            Desired line thickness for box outline. 
        """
        draw(image_path, out_dir, draw_type='box_3d', object_ids=object_ids, object_types=object_types, line_thickness=line_thickness)


    def segmentation(self, image_path, out_dir, object_ids=None, object_types=None, line_thickness=1):
        """
        Generates images annotated with outlines around objects for datasets downloaded from the Platform. Optional filter on 
        object_ids or object_types (must choose a single filter).
        
        Parameters
        ----------
        image_path : str
            Path to of specific image file to draw the boxes for.  
        out_dir : str
            File path to directory where the image should be saved to.
        object_ids : list[int]
            List of object id's to annotate. If not provided, all objects will get annotated. Choose either id or type filter.
        object_types: list[str]
            Filter for the object types to annotate. If not provided, all object types will get annotated. Choose either id or type filter.
        line_thickness: int
            Desired line thickness for object segmentation outline. 
        """
        draw(image_path, out_dir, draw_type='segmentation', object_ids=object_ids, object_types=object_types, line_thickness=line_thickness)


    def dump_coco(self, datadir, outdir, mapfile):
        """Generates annotations in the format of COCO Object Detection. See https://cocodataset.org/#format-data.
        
        Parameters
        ----------
        datadir : str
            The location of the Ana dataset.
        outdir : str
            The location to output the annotation files to.
        mapfile: str
            The location of the mapping file.
        """
        with open(mapfile) as f:
            mapping = yaml.safe_load(f)
        convert_coco(datadir, outdir, mapping)


    def dump_kitti(self, datadir, outdir, mapfile):
        """Generates annotations in the format of KITTI. See https://docs.nvidia.com/metropolis/TLT/archive/tlt-20/tlt-user-guide/text/preparing_data_input.html.
        
        Parameters
        ----------
        datadir : str
            The location of the Ana dataset.
        outdir : str
            The location to output the annotation files to.
        mapfile: str
            The location of the mapping file.
        """
        with open(mapfile) as f:
            mapping = yaml.safe_load(f)
        convert_kitti(datadir, outdir, mapping)


    def dump_pascal(self, datadir, outdir, mapfile):
        """Generates annotations in the format of PASCAL VOC. See https://pjreddie.com/media/files/VOC2012_doc.pdf.
        
        Parameters
        ----------
        datadir : str
            The location of the Ana dataset.
        outdir : str
            The location to output the annotation files to.
        mapfile: str
            The location of the mapping file.
        """
        with open(mapfile) as f:
            mapping = yaml.safe_load(f)
        convert_pascal(datadir, outdir, mapping)


    def dump_sagemaker_od(self, datadir, outdir, mapfile):
        """Generates annotations in the format of Sagemaker Object Detection. See https://docs.aws.amazon.com/sagemaker/latest/dg/object-detection.html.
        
        Parameters
        ----------
        datadir : str
            The location of the Ana dataset.
        outdir : str
            The location to output the annotation files to.
        mapfile: str
            The location of the mapping file.
        """
        with open(mapfile) as f:
            mapping = yaml.safe_load(f)
        convert_sagemaker_od(datadir, outdir, mapping)


    def dump_sagemaker_ss(self, datadir, outdir, mapfile):
        """Generates annotations in the format of Sagemaker Semantic Segmentation. See https://docs.aws.amazon.com/sagemaker/latest/dg/semantic-segmentation.html.
        
        Parameters
        ----------
        datadir : str
            The location of the Ana dataset.
        outdir : str
            The location to output the annotation files to.
        mapfile: str
            The location of the mapping file.
        """
        with open(mapfile) as f:
            mapping = yaml.safe_load(f)
        convert_sagemaker_ss(datadir, outdir, mapping)



"""
Volumes Functions
"""

def get_volumes(self, volumeId=None, organizationId=None):
    """Retrieves all volumes the user has access to.
    
    Parameters
    ----------
    volumeId : str
        The ID of a specific Volume.
    organizationId : str
        The ID of the organization that the volume belongs to.
    
    Returns
    -------
    list[dict]
        Volume Info
    """
    if self.check_logout(): return
    volumes = self.ana_api.getVolumes(organizationId=organizationId, volumeId=volumeId)
    if volumes:
        for volume in volumes:
            self.volumes[volume['volumeId']] = volume['name']
    return volumes
    


def get_managed_volumes(self, volumeId=None, organizationId=None):
    """Retrieves the managed volumes for an organization.
    
    Parameters
    ----------
    volumeId : str
        The ID of a specific Volume.
    organizationId : str
        The ID of the organization that the managed volume belongs to.
    
    Returns
    -------
    list[dict]
        Volume Info
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getManagedVolumes(organizationId=organizationId, volumeId=volumeId)


def create_managed_volume(self, name, organizationId=None):
    """Creates a new volume with the specified name in the organization.
    
    Parameters
    ----------
    name : str
        The name of the new volume. Note: this name needs to be unique per organization.
    organizationId : str
        The ID of the organization that the managed volume will belong to.
    
    Returns
    -------
    str
        volumeId
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    if name is None: raise Exception("Name must be specified.")
    return self.ana_api.createManagedVolume(organizationId=organizationId,name=name)

    
def delete_managed_volume(self, volumeId):
    """Removes the volumeId from the organization. Note that this will delete any remote data in the volume 
    and channels that rely on this volume will need to be updated.
    
    Parameters
    ----------
    volumeId : str
        The ID of a specific Volume to delete.
    
    Returns
    -------
    str
        Status
    """
    if self.check_logout(): return
    if volumeId is None: raise Exception('VolumeId must be specified.')
    return self.ana_api.deleteManagedVolume(volumeId=volumeId)


def edit_managed_volume(self, volumeId, name):
    """Edits the name of a volume.
    
    Parameters
    ----------
    volumeId: str
        The volumeId that will be updated.
    name : str
        The new name of the new volume. Note: this name needs to be unique per organization.
    
    Returns
    -------
    str
        Status
    """
    if self.check_logout(): return
    if volumeId is None: raise Exception('VolumeId must be specified.')
    if name is None: raise Exception("Name must be specified.")
    return self.ana_api.editManagedVolume(volumeId=volumeId,name=name)


def add_volume_access(self, volumeId, organizationId):
    """Add access to a volume for an organization.
    
    Parameters
    ----------
    volumeId : str
        VolumeId to add access for.
    organizationId : str
        Organization ID to add access
    
    Returns
    -------
    str
        Status
    """
    if self.check_logout(): return
    if organizationId is None: raise Exception('OrganizationId must be specified.')
    if volumeId is None: raise Exception('VolumeId must be specified.')
    return self.ana_api.addVolumeOrganization(volumeId=volumeId, organizationId=organizationId)


def remove_volume_access(self, volumeId, organizationId):
    """Remove access to a volume for an organization.
    
    Parameters
    ----------
    volumeId : str
       VolumeId to remove access to.
    organizationId : str
        Organization ID to remove access.
    
    Returns
    -------
    str
       Status
    """
    if self.check_logout(): return
    if organizationId is None: raise Exception('OrganizationId must be specified.')
    if volumeId is None: raise Exception('VolumeId must be specified.')
    return self.ana_api.removeVolumeOrganization(volumeId=volumeId, organizationId=organizationId)


def get_volume_data(self, volumeId, files=[], dir=None):
    """Retrieves information about data from a volume.
    
    Parameters
    ----------
    volumeId : str
       VolumeId to remove access to.
    files : str
        The specific files or directories to retrieve information about from the volume, if you wish to retrieve all then leave the list empty.
    dir : str
        Specific volume directory to retrieve information about. Optional. 
    Returns
    -------
    str
       Status
    """
    if self.check_logout(): return
    if volumeId is None: raise Exception('VolumeId must be specified.')
    response = self.ana_api.getVolumeData(volumeId=volumeId, keys=files, dir=dir)
    result = []
    for fileinfo in response:
        result.append({
            'key':          fileinfo['key'],
            'size':         fileinfo['size'],
            'lastUpdated':  fileinfo['updatedAt'],
            'hash':         fileinfo['hash']
        })
    return result


def download_volume_data(self, volumeId, files=[], localDir=None, sync=False):
    """Download data from a volume.
    
    Parameters
    ----------
    volumeId : str
       VolumeId to download data of.
    files : str
        The specific files or directories to retrieve from the volume, if you wish to retrieve all then leave the list empty.
    localDir : str
        The location of the local directory to download the files to. If not specified, this will download the files to the current directory.
    sync: bool
        Recursively downloads new and updated files from the source to the destination. Only creates folders in the destination if they contain one or more files.
    Returns
    -------
    str
       Status
    """
    import hashlib, requests, traceback, os
    if self.check_logout(): return
    if volumeId is None: raise Exception('VolumeId must be specified.')
    if localDir is None: localDir = os.getcwd()
    if not os.path.exists(localDir): raise Exception(f"Could not find directory {localDir}.")

    response = self.ana_api.getVolumeData(volumeId=volumeId, keys=files)
    source_hashes = list(map((lambda x: x['key'] + x['hash']), response))
    destination_files = []
    destination_hashes = []

    if sync == True:    
        for root, dirs, files in os.walk(localDir):
            for file in files:
                filepath = os.path.join(root, file).replace(localDir, '')
                destination_files.append(filepath)
                file_hash = hashlib.md5()
                with open(os.path.join(root, file),'rb') as f: 
                    while True:
                        chunk = f.read(128 * file_hash.block_size)
                        if not chunk:
                            break
                        file_hash.update(chunk)
                destination_hashes.append(filepath + file_hash.hexdigest())

    for index, hash in enumerate(source_hashes):
        if (sync == True and (hash in destination_hashes)):
            print(f"\x1b[1K\rsync: {response[index]['key']}'s hash exists in {localDir}", end='\n' if self.verbose else '', flush=True)
        elif sync == False or (hash not in destination_hashes):
            try:
                downloadresponse = requests.get(url=response[index]['url'])
                filename = os.path.join(localDir, response[index]['key'])
                if not os.path.exists(os.path.dirname(filename)):
                    os.makedirs(os.path.dirname(filename))
                with open(filename, 'wb') as outfile:
                    outfile.write(downloadresponse.content)
                print(f"\x1b[1K\rdownload: {response[index]['key']} to {filename}", end='\n' if self.verbose else '', flush=True)
            except:
                traceback.print_exc()
                print(f"\x1b[1K\rdownload: failed to download {response[index]['key']}", end='\n' if self.verbose else '', flush=True)

    return


def upload_volume_data(self, volumeId, files=[], localDir=None, sync=False):
    """Upload data to a volume.
    
    Parameters
    ----------
    volumeId : str
       VolumeId to upload data to.
    files : list[str]
        The specific files or directories to push to the volume from the localDir. If you wish to push all data in the root directory, then leave the list empty.
    localDir : str
        The location of the local directory to upload the files from. If not specified, this will try to upload the files from the current directory.
    sync: bool
        Recursively uploads new and updated files from the source to the destination. Only creates folders in the destination if they contain one or more files.
    Returns
    -------
    str
       Status
    """
    import hashlib, requests, traceback, time, os

    if self.check_logout(): return
    if volumeId is None: raise Exception('VolumeId must be specified.')
    if localDir is None: localDir = os.getcwd()
    if not localDir.endswith('/'): localDir+='/'
    if not os.path.exists(localDir): raise Exception(f"Could not find directory {localDir}.")

    source_files = []
    source_hashes = []
    faileduploads = []
        
    if len(files):
        for file in files:
            filepath = os.path.join(localDir, file)
            if os.path.isdir(filepath):
                for root, dirs, files in os.walk(filepath):
                    for file in files:
                        filepath = os.path.join(root, file).replace(localDir, '')
                        source_files.append(filepath)
                        if sync == True:
                            file_hash = hashlib.md5()
                            with open(os.path.join(root,file),'rb') as f: 
                                while True:
                                    chunk = f.read(128 * file_hash.block_size)
                                    if not chunk: break
                                    file_hash.update(chunk)
                            source_hashes.append(filepath + file_hash.hexdigest())
            elif os.path.isfile(filepath):
                source_files.append(file)
                if sync == True:
                    file_hash = hashlib.md5()
                    with open(filepath,'rb') as f: 
                        while True:
                            chunk = f.read(128 * file_hash.block_size)
                            if not chunk: break
                            file_hash.update(chunk)
                    source_hashes.append(file + file_hash.hexdigest())
            else: print(f"Could not find {filepath}.")
    else:
        for root, dirs, files in os.walk(localDir):
            for file in files:
                filepath = os.path.join(root, file).replace(localDir, '')
                source_files.append(filepath)
                if sync == True:
                    file_hash = hashlib.md5()
                    with open(os.path.join(root,file),'rb') as f: 
                        while True:
                            chunk = f.read(128 * file_hash.block_size)
                            if not chunk:
                                break
                            file_hash.update(chunk)
                    source_hashes.append(filepath + file_hash.hexdigest())

    if sync == True:
        response = self.ana_api.getVolumeData(volumeId=volumeId, keys=files)
        destination_hashes = list(map((lambda x: x['key'] + x['hash']), response))
        delete_files = []
        for index, object in enumerate(response):
            if object['key'] not in source_files:
                delete_files.append(object['key'])  

        if (len(delete_files)):
            print(f"The following files will be deleted:", end='\n', flush=True)
            for file in delete_files:
                print(f"   {file}", end='\n', flush=True)
            answer = input("Delete these files [Y/n]: ")
            if answer.lower() == "y":
                self.refresh_token()
                self.ana_api.deleteVolumeData(volumeId=volumeId, keys=delete_files)

    for index, file in enumerate(source_files):
        if (sync == True and (source_hashes[index] in destination_hashes)):
            print(f"\x1b[1K\rsync: {file}'s hash exists", end='\n' if self.verbose else '', flush=True)
        elif sync == False or (source_hashes[index] not in destination_hashes):
            self.refresh_token()
            fileinfo = self.ana_api.putVolumeData(volumeId=volumeId, keys=[file])[0]
            try:
                filepath = os.path.join(localDir, file)
                
                print(f"\x1b[1K\rupload: {file} to the volume. [{index+1} / {len(source_files)}]", end='\n' if self.verbose else '', flush=True)
                with open(filepath, 'rb') as filebytes:
                    files = {'file': filebytes}
                    data = {
                        "key":                  fileinfo['fields']['key'],
                        "bucket":               fileinfo['fields']['bucket'],
                        "X-Amz-Algorithm":      fileinfo['fields']['algorithm'],
                        "X-Amz-Credential":     fileinfo['fields']['credential'],
                        "X-Amz-Date":           fileinfo['fields']['date'],
                        "X-Amz-Security-Token": fileinfo['fields']['token'],
                        "Policy":               fileinfo['fields']['policy'],
                        "X-Amz-Signature":      fileinfo['fields']['signature'],
                    }
                    response = requests.post(fileinfo['url'], data=data, files=files)
                    if response.status_code != 204:
                        faileduploads.append(file)
                        print(f"\x1b[1K\rupload: {file} failed", end='\n' if self.verbose else '', flush=True)
                    else:
                        print(f"\x1b[1K\rupload: {file} completed", end='\n' if self.verbose else '', flush=True)
            except:
                traceback.print_exc()
                faileduploads.append(file)
                print(f"\x1b[1K\rupload: {file} failed", end='\n' if self.verbose else '', flush=True)
    print("\x1b[1K\rUploading files completed.", flush=True)
    if len(faileduploads): print('The following files failed to upload:', faileduploads, flush=True)
    return
            

def delete_volume_data(self, volumeId, files=[]):
    """Delete data from a volume.
    
    Parameters
    ----------
    volumeId : str
       VolumeId to delete files from.
    files : str
        The specific files to delete from the volume. If left empty, no files are deleted.
    
    Returns
    -------
    str
       Status
    """
    if self.check_logout(): return
    if volumeId is None: raise Exception('VolumeId must be specified.')
    return self.ana_api.deleteVolumeData(volumeId=volumeId, keys=files)


def mount_volumes(self, volumes):
    """Retrieves credentials for mounting volumes.
    
    Parameters
    ----------
    volumes : [str]
       Volumes to retrieve mount credentials for.

    Returns
    -------
    dict
        Credential information.
    """
    if self.check_logout(): return
    if not len(volumes): raise Exception('A list of volumeIds must be specified.')
    return self.ana_api.mountVolumes(volumes=volumes)



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import ast
import sys
import anatools.lib.context as ctx
from anatools.lib.node import Node

logger = logging.getLogger(__name__)


def _float_or_list(val):
    """Input conversion - expecting a float or a list of floats"""
    if isinstance(val, str):
        try:
            val = ast.literal_eval(val)
        except:
            logger.error("Expected float or list of floats")
            sys.exit(1)
    if isinstance(val, list) or isinstance(val, float):
        return val
    elif isinstance(val, int):
        return float(val)
    else:
        logger.error("Expected float or list of floats")
        sys.exit(1)

def _int_or_list(val):
    """Input conversion - expecting an int or a list of ints"""
    if isinstance(val, str):
        try:
            val = ast.literal_eval(val)
        except:
            logger.error("Expected integer or list of integers")
            sys.exit(1)
    if isinstance(val, list) or isinstance(val, int):
        return val
    else:
        logger.error("Expected integer or list of integers")
        sys.exit(1)

def _none_or_int_or_list(val):
    """Input conversion - expecting None, empty string, int, or a list of ints"""
    if val is None or val == "":
        # assume empty string means null - necessary for GUI when user backspaces over field
        return None
    elif isinstance(val, str):
        try:
            val = ast.literal_eval(val)
        except:
            logger.error("Expected empty string, integer, or list of integers")
            sys.exit(1)
    if isinstance(val, list):
        return tuple(val)
    elif isinstance(val, int):
        return val
    else:
        logger.error("Expected empty string, integer, or list of integers")
        sys.exit(1)

class RandomTriangular(Node):
    """
    Draw random samples from a triangular distribution over the closed interval [left, right].
    See numpy.random.triangular for details.
    """

    def exec(self):
        """Execute node"""
        logger.info("Executing {}".format(self.name))

        # get inputs
        left = _float_or_list(self.inputs["left"][0])
        mode = _float_or_list(self.inputs["mode"][0])
        right = _float_or_list(self.inputs["right"][0])
        size = _none_or_int_or_list(self.inputs["size"][0])

        # draw samples
        out = ctx.random.triangular(left, mode, right, size)

        return {"out": out}

class RandomUniform(Node):
    """
    Draw random samples from a uniform distribution over the half-open interval [low, high).
    See numpy.random.uniform for details.
    """

    def exec(self):
        """Execute node"""
        logger.info("Executing {}".format(self.name))

        # get inputs
        low = _float_or_list(self.inputs["low"][0])
        high = _float_or_list(self.inputs["high"][0])
        size = _none_or_int_or_list(self.inputs["size"][0])

        # draw samples
        out = ctx.random.uniform(low, high, size)

        return {"out": out}

class RandomRandint(Node):
    """
    Return random integers from low (inclusive) to high (exclusive).
    See numpy.random.randint for details.
    """

    def exec(self):
        """Execute node"""
        logger.info("Executing {}".format(self.name))

        # get inputs
        low = _int_or_list(self.inputs["low"][0])
        high = _int_or_list(self.inputs["high"][0])
        size = _none_or_int_or_list(self.inputs["size"][0])

        # draw samples
        out = ctx.random.randint(low, high, size)

        return {"out": out}

class RandomNormal(Node):
    """
    Draw random samples from a normal (Gaussian) distribution.
    See numpy.random.normal for details.
    """

    def exec(self):
        """Execute node"""
        logger.info("Executing {}".format(self.name))

        # get inputs
        loc = _float_or_list(self.inputs["loc"][0])
        scale = _float_or_list(self.inputs["scale"][0])
        size = _none_or_int_or_list(self.inputs["size"][0])

        # draw samples
        out = ctx.random.normal(loc, scale, size)

        return {"out": out}


class RandomChoice(Node):

    def exec(self):
        logger.info("Executing {}".format(self.name))
        # parse inputs
        choice_list = list(self.inputs["List_of_Choices"][0])
        number = int(self.inputs["Number_of_Choices"][0])
        unique = str(self.inputs["Unique_Choices"][0])
        if unique is "True":    choices = ctx.random.choice(choice_list, number, replace=False) 
        else:                   choices = ctx.random.choice(choice_list, number, replace=True) 
        return {"Choices": choices}



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from abc import ABC
import logging
import bpy
from anatools.lib.search_utils import find_root
import anatools.lib.bbox as annotations

logger = logging.getLogger(__name__)


class AnaObject(ABC):
    """ Base class for Ana objects """

    next_instance = 1

    def __init__(self, object_type):

        # instance number
        self.instance = AnaObject.next_instance
        AnaObject.next_instance += 1
        # object type
        self.object_type = object_type
        # loaded flag is set to true after the object has been loaded into the current scene
        self.loaded = False
        # pointer to bpy.data.collections["collectionName"] where collectionName is the collection containing the object
        self.collection = None # set in loader
        # pointer to bpy.data.objects["objectName"] where objectName is the name of the root object
        self.root = None # set in loader
        # does this object appear in the rendered image? This will be set in the annotation module
        self.rendered = False
        # modifier metadata
        self.modifiers = []
        # object of interest - set when adding object to scene
        self.ooi = False
        # location of mask file for scene rendered with only this object
        self.solo_mask_id = ''
        # object specific configuration
        self.config = {}

    def __key(self):
        # the key for hashing and equality comparison
        return (self.instance,)

    def __hash__(self):
        # make the object hashable so we can use it as a dictionary key
        return hash(self.__key())

    def load(self, **kwargs):
        """
        Default loader - Load the object from a file

        If you override this method in your subclass you must do all of the following
            - create the object hierarchy with a single root object
            - create the object collection
            - link all objects to the collection
            - link the collection to the current scene
            - set self.collection = the collection datablock
            - set self.root = the root object datablock
            - set self.loaded = True
       """
        if self.loaded:
            # only load the object once
            return

        blender_file = kwargs.pop("blender_file")
        # load the collection
        with bpy.data.libraries.load(filepath="//" + blender_file, link=False) as (_, dt):
            dt.collections = [self.object_type]
        self.collection = dt.collections[0]
        
        # link collection to the current scene
        bpy.context.scene.collection.children.link(self.collection)

        # find the root object
        self.root = find_root(self.collection)
        self.loaded = True

        # save object config if it was provided
        if "config" in kwargs:
            self.config = kwargs.pop("config")


    def dump_metadata(self):
        """
        Convert object to a JSON serializable representation.
        Sub classes add metadata by implementing the dump_metadata method as follows:
            def dump_metadata(self):
                meta = super().dump_metadata()
                return {**meta, "subclass-attribute": subclass-value, ...}
        """
        metadata = {
            "id": self.instance,
            "type": self.object_type
        }

        if len(self.modifiers) > 0:
            metadata["modifiers"] = self.modifiers
        
        return metadata

    def dump_annotations(self, calculate_obstruction=False):
        """ Generate annotations for the object. """
        if not self.ooi:
            return
        seg, bbox = annotations.compute_polygons(self)
        if seg is None or bbox is None:
            self.rendered = False
            return
        bbox3d = annotations.compute_bbox3d(self)
        centroid, distance = annotations.compute_centroid(self)
        truncated = annotations.truncated(self, bbox)
        if calculate_obstruction:
            obstruction = annotations.compute_obstruction(self)
        else:
            obstruction = None
        size = annotations.compute_size(self)
        rotation = annotations.compute_rotation(self)
        annotation = {
            'id':           self.instance,
            'bbox':         bbox,
            'segmentation': seg,
            'bbox3d':       bbox3d,
            'centroid':     centroid,
            'distance':     distance,
            'truncated':    truncated,
            'size':         size,
            'rotation':     rotation,
            'obstruction':  obstruction}
        return annotation

    def find_object(self, qname):
        """
        Recursively search the object hierarchy for a child object where
        qname is a list of object base names from the original hierarchy.

        For example, given this object hierarchy in the original blender file:

        Aircraft
          Left_Wing
            Flap
          Right_Wing
            Flap.001

        To find the right wing flap in a loaded instance of this object you
        would do the following, assuming "aircraft" is the AnaObject instance.

        obj = aircraft.find_object(["Right_Wing", "Flap"])

        IMPORTANT CAVEAT:

        This will not work if an object in the hierarchy has two children with
        the same base name, e.g.

        Aircraft
          Left_Wing
            Flap
            Flap.001
          Right_Wing
            Flap.002
            Flap.003

        To solve this problem you need to modify the original file to follow the
        "no two children with the same base name" rule

        """
        
        # start at the root
        obj = self.root
        # find objects below the root
        for level, name in enumerate(qname):
            found_child = False
            # if exact match not found then the object was renamed when loaded
            for child in obj.children:
                if child.name.startswith(name):
                    found_child = True
                    obj = child
                    break
            if not found_child:
                # no child matched qname at this level
                logger.critical("Ana object '%s' does not contain '%s'", self.root.name, " ".join(qname[:level+1]))
                sys.exit(1)
        # if all levels matched then we found the object
        return obj



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import numpy
from anatools.lib.node       import Node
from anatools.lib.generator  import CreateBranchGenerator

logger = logging.getLogger(__name__)


class SelectGenerator(Node):
    def exec(self):
        """ Select a generator from a list of generators. """
        logger.info("Executing {}".format(self.name))
        try:
            children = numpy.array(self.inputs["Generators"]).flatten()
            generator = CreateBranchGenerator(children=children)
        except Exception as e:
            logger.error("{} in \"{}\": \"{}\"".format(type(e).__name__, type(self).__name__, e).replace("\n", ""))
            raise
        return {"Generator": generator} 


class Weight(Node):
    """ Add modify the weight of a generator/modifier """
    def exec(self):
        logger.info("Executing {}".format(self.name))

        if len(self.inputs["Generator"]) != 1:
            logger.error("Weight 'generator' input port requires exactly 1 link")
            raise ValueError
        generator = self.inputs["Generator"][0]
        try:
            generator.weight = float(self.inputs["Weight"][0])
        except Exception as e:
            logger.error("{} in \"{}\": \"{}\"".format(type(e).__name__, type(self).__name__, e).replace("\n", ""))
            raise
        return {"Generator": generator}


class SetInstanceCount(Node):
    """ Set the count property of a generator to produce multiple instances. """
    def exec(self):
        logger.info("Executing {}".format(self.name))

        if len(self.inputs["Generator"]) != 1:
            logger.error("SetInstanceCount 'generator' port requires exactly 1 link.")
            raise ValueError
        generator = self.inputs["Generator"][0]
        try:
            generator.count = int(self.inputs["Count"][0])
        except Exception as e:
            logger.error("{} in \"{}\": \"{}\"".format(type(e).__name__, type(self).__name__, e).replace("\n", ""))
            raise
        return {"Generator": generator}



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import os
import argparse
import yaml
import logging
from anatools.lib.channel import Channel
from anatools.lib.interp import interp

logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser()
parser.add_argument('--channel', required=True)
parser.add_argument('--graph', required=True)
parser.add_argument('--loglevel')
parser.add_argument('--logfile')
parser.add_argument('--seed', type=int)
parser.add_argument('--interp_num', type=int)
parser.add_argument('--preview', action="store_true")
parser.add_argument('--output')
parser.add_argument('--data')
args = parser.parse_args()

# Configure initial logging (this may get overridden later)
Channel.configure_logging(loglevel=args.loglevel, logfile=args.logfile, logfile_mode="a")

# get channel file name
channel_file = args.channel
if os.path.splitext(channel_file)[1] == '':
    channel_file = channel_file + ".yml"

channel = Channel(channel_file)
channel.process_args(args)

# read graph
with open(args.graph, "r") as f:
    input_graph = yaml.safe_load(f)

# execute setup
channel.setup()

# interpret graph
try:
    interp(input_graph)
except Exception as e:
    message = f"An exception of type {type(e).__name__} occurred while interpreting graph"
    logging.error(message, exc_info=e)
    sys.exit(1)


import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as rf:
    requirements = rf.read().splitlines()

def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + '/__init__.py').read())
    return result.group(1)

setuptools.setup(
    name="anatools",
    version=get_property('__version__', 'anatools'),
    author="Rendered AI, Inc",
    author_email="support@rendered.ai",
    description="Tools for development with the Rendered.ai Platform.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://rendered.ai",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    scripts=[
        "anatools/bin/ana", 
        "anatools/bin/anadeploy", 
        "anatools/bin/anamount",
        "anatools/bin/anautils"],
    package_data={"": ["*.yml"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"],
    python_requires='>=3.6',
)



"""
GAN Functions
"""
import os
import requests

def get_gan_models(self, modelId=None, workspaceId=None):
    """Retrieve information about GAN models that exist on the platform.
    
    Parameters
    ----------
    modelId : str
        Model ID to retrieve information for. 
    
    Returns
    -------
    list[dict]
        GAN Model information.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getGANModels(modelId=modelId, workspaceId=workspaceId)
    

def get_gan_dataset(self, datasetId, workspaceId=None):
    """Retrieve information about GAN dataset jobs.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to retrieve information for. 
    workspaceId : str
        Workspace ID where the dataset exists.
    
    Returns
    -------
    list[dict]
        Information about the GAN Dataset.
    """
    if self.check_logout(): return
    if datasetId is None: raise ValueError("DatasetId must be provided.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getGANDataset(workspaceId=workspaceId, datasetId=datasetId)


def create_gan_dataset(self, modelId, datasetId, workspaceId=None):
    """Create a new GAN dataset based off an existing dataset. This will start a new job.
    
    Parameters
    ----------
    modelId : str
        Model ID to use for the GAN.
    datasetId : str
        Dataset ID to input into the GAN. 
    workspaceId : str
        Workspace ID where the dataset exists.
    
    Returns
    -------
    str
        The datsetId for the GAN Dataset job.
    """
    if self.check_logout(): return
    if modelId is None: raise ValueError("ModelId must be provided.")
    if datasetId is None: raise ValueError("DatasetId must be provided.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.createGANDataset(workspaceId=workspaceId, datasetId=datasetId, modelId=modelId)


def delete_gan_dataset(self, datasetId, workspaceId=None):
    """Deletes a GAN dataset job.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID for the GAN dataset. 
    workspaceId : str
        Workspace ID where the dataset exists.
    
    Returns
    -------
    bool
        Returns true if the GAN dataset was successfully deleted.
    """
    if self.check_logout(): return
    if datasetId is None: raise ValueError("DatasetId must be provided.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteGANDataset(workspaceId=workspaceId, datasetId=datasetId)


def upload_gan_model(self, modelFileName, name, description, flags=None, organizationId=None):
    """Uploades a GAN model to the microservice. The model will be owned by the specified organization.
    If not organizationId is given the model will be owned by that of the analcient.
    
    Parameters
    ----------
    modelFileName : str
        The file of the model - relative to the local directry.
    name : str
        A name for model.
    description : str
        Details about the model.
    flags : str
        Parameters for use when running the model.
    organizationId : str
        Id of organization that owns the model, that of the anaclient if not given.
    
    Returns
    -------
    modleId : str
        The unique identifier for this model.
    """
    if self.check_logout(): return

    if not os.path.exists(modelFileName):
        print("File not found in " + os.getcwd())
        return

    if organizationId is None: organizationId = self.organization

    self.refresh_token()
    fileinfo = self.ana_api.uploadGANModel(organizationId=organizationId, name=name, description=description, flags=flags)
    # fileinfo keys:
    # "key": S3 Key
    # "modelId": modelId,
    # "url": s3 url
    # "fields": dictionary of details to access presigned url
    if not fileinfo:
        print(fileinfo)
        return
    
    try:
        with open(modelFileName, 'rb') as filebytes:
            files = {'file': filebytes}
            data = {
                "key":                  fileinfo['fields']['key'],
                "bucket":               fileinfo['fields']['bucket'],
                "X-Amz-Algorithm":      fileinfo['fields']['algorithm'],
                "X-Amz-Credential":     fileinfo['fields']['credential'],
                "X-Amz-Date":           fileinfo['fields']['date'],
                "X-Amz-Security-Token": fileinfo['fields']['token'],
                "Policy":               fileinfo['fields']['policy'],
                "X-Amz-Signature":      fileinfo['fields']['signature'],
            }
            response = requests.post(fileinfo['url'], data=data, files=files)
            if response.status_code != 204:
                if self.verbose: print(f"Failure", flush=True)
            else:
                if self.verbose: print('Success', flush=True)
    except Exception as e:
        # traceback.print_exc()
        print('Failed to upload: {}'.format(e), flush=True)

    return fileinfo['modelId']


def delete_gan_model(self, modelId):
    """Delete the GAN model and remove access to it from all shared organizations.
    This can only be done by a user in the organization that owns the model.
    
    Parameters
    ----------
    modelId : str
        The ID of a specific GAN model.
    
    Returns
    -------
    str
        Status
    """
    if self.check_logout(): return
    if modelId is None: raise Exception('ModelId must be specified.')
    return self.ana_api.deleteGANModel(modelId=modelId)


def add_gan_access(self, modelId, targetOrganizationId):
    """Adds access to a model for new organization; can only be done by a user in the organization that owns the model.
    
    Parameters
    ----------
    modelId : str
        Id for the GAN model.
    targetOrganizationId : str
        Organization ID where the model is be granted access.
    
    Returns
    -------
    bool
        Returns true if the GAN model was granted access.
    """
    if self.check_logout(): return
    if modelId is None: raise ValueError("Model id must be provided.")
    if targetOrganizationId is None: raise ValueError("Target organization id must be provided.")
    return self.ana_api.addGANAccess(organizationId=targetOrganizationId, modelId=modelId)


def remove_gan_access(self, modelId, targetOrganizationId):
    """Removes access to a model for an that does not own the model.
    
    Parameters
    ----------
    modelId : str
        Id for the GAN model.
    targetOrganizationId : str
        Organization ID where the model access is to be removed.
    
    Returns
    -------
    bool
        Returns true if the GAN model access was removed.
    """
    if self.check_logout(): return
    if modelId is None: raise ValueError("Model id must be provided.")
    if targetOrganizationId is None: raise ValueError("Target organization id must be provided.")
    return self.ana_api.removeGANAccess(organizationId=targetOrganizationId, modelId=modelId)


import os
import json
import yaml
import numpy as np
from PIL import Image


def get_image_size(metadata, imagesdir):
    """ Helper function to get properties of an image"""
    if 'sensor' in metadata and 'resolution' in metadata['sensor']:
        imagesize = metadata['sensor']['resolution']
    else:
        image = Image.open(os.path.join(imagesdir, metadata['filename']))
        imagesize = image.size
    return imagesize


def kitti_label(cat, objann, image_size, metadata):
    """ Create a KITTI label for an object.
    # Support perspective / rectilinear projections as well as orthographic projections

    Kitti label format:
    - https://docs.nvidia.com/metropolis/TLT/archive/tlt-20/tlt-user-guide/text/preparing_data_input.html
    - https://github.com/bostondiditeam/kitti/blob/master/resources/devkit_object/readme.txt
    """
    # Commonly used inputs
    sensor_metadata = metadata['sensor']
    lens_type = 'perspective' # default lens type to perspective
    if 'lens_type' in sensor_metadata: lens_type = sensor_metadata['lens_type']

    # Build the label array
    label = list()

    # Category name: str (1 entry)
    label.append(cat.lower().replace(' ', ''))  # NVIDIA models require lowercase names

    # Is truncated: boolean int (1 entry)
    truncated = '0'  # Assuming object does not leave image boundary
    if 'truncated' in objann: truncated = '1'
    label.append(truncated)

    # Occluded level (fully visible, largely visible, largely occluded, unk): int in the set [0,1,2,3] (1 entry)
    occlusionLevel = '3'
    if 'obstruction' in objann:
        if objann['obstruction'] is None:   occlusionLevel = '3'
        elif objann['obstruction'] == 0.0:  occlusionLevel = '0'
        elif objann['obstruction'] < 0.5:   occlusionLevel = '1'
        else:                               occlusionLevel = '2'
    label.append(occlusionLevel)

    # Observation angle: float [-pi..pi] (1 entry)
    centroid = np.array([objann['centroid'][1], objann['centroid'][0]])
    imageCenter = np.array(image_size) / 2

    if lens_type == 'perspective' and 'camera_size' in sensor_metadata and 'focal_length' in sensor_metadata:
        # https://en.wikipedia.org/wiki/Angle_of_view
        angle_of_view = 2 * np.arctan(sensor_metadata['camera_size'] / (2 * sensor_metadata['focal_length']))
        # hfov = 2*sensor_metadata['target_distance']*np.tan(angle_of_view/2)
        # px_per_distance = image_size[0]/hfov  # pixels per mm

        r_px = np.linalg.norm(centroid - imageCenter)
        alpha = r_px / image_size[0] * angle_of_view  #linear model of full angle

    elif 'gsd_at_nadir' in sensor_metadata:
        distance_from_sensor = objann['distance']  # TODO check lens_orthographic_scale

        if sensor_metadata['look_angle']:
            raise # TODO: assuming sensor is in center of image pointing straight down
        if sensor_metadata['azimuth'] != 180:
            raise # TODO: assuming pixel grid is aligned with the camera reference frame

        distance_from_center = np.linalg.norm(centroid-imageCenter)*sensor_metadata['gsd_at_nadir']
        alpha = np.arccos(distance_from_center/distance_from_sensor)
    else:
        alpha = 0

    label.append(str(alpha))

    # bbox coords [xmin, ymin, xmax, ymax]: integers (4 entries)
    xmin = objann['bbox'][0]
    ymin = objann['bbox'][1]
    xmax = xmin + objann['bbox'][2]
    ymax = ymin + objann['bbox'][3]
    label.extend([str(xmin), str(ymin), str(xmax), str(ymax)])

    # 3-D dimension of the object (in real distance units): floats (3 entries)
    length, width, height = objann.get('size', [1.0, 1.0, 1.0])
    label.extend([str(height), str(width), str(length)])

    # 3-D object location x, y, z in camera coordinates (in real distance units): floats (3 entries)
    if lens_type == 'perspective':
        distance_from_sensor = objann['distance']
        r = distance_from_sensor * np.sin(alpha)
        z = distance_from_sensor * np.cos(alpha)

        deltax_px = centroid[0] - imageCenter[0]
        deltay_px = centroid[1] - imageCenter[1]
        if deltax_px == 0:
            x = 0
            y = r
        else:
            theta = np.arctan(deltay_px/deltax_px)
            x = r * np.sin(theta)
            y = r * np.cos(theta)

    elif 'gsd_at_nadir' in sensor_metadata:  # Orthographic Projection
        x = (centroid[0]-imageCenter[0])*sensor_metadata['gsd_at_nadir']
        y = (centroid[1]-imageCenter[1])*sensor_metadata['gsd_at_nadir']
        #z = objann['distance'] * #TODO check lens_orthographic_scale
        z = 36000000.0
    else:
        x,y,z = 0,0,0

    label.extend([str(x), str(y), str(z)])

    # Rotation ry about the Y-axis in camera coordinates: float [-pi..pi] (1 entry)
    # In Euler angles this would be the 'yaw' as y-axiz in camera coords is z-axis in Euler math (and in Blender)
    rotation = objann.get('rotation', [0.0, 0.0, 0.0])
    label.append(str(rotation[2]))

    return ' '.join(label)


def convert_kitti(datadir, outdir, mapping):
    """ To use the data in KITTI format. Result will be placed in outdir
    """
    # Input directories
    imagesdir = os.path.join(datadir, 'images')
    annsdir = os.path.join(datadir, 'annotations')
    metadir = os.path.join(datadir, 'metadata')


    # for each interpretation, gather annotations and map categories
    for annsfilename in os.listdir(annsdir):
        with open(os.path.join(annsdir, annsfilename), 'r') as af:
            anns = json.load(af)
        with open(os.path.join(metadir, annsfilename.replace('ana', 'metadata')), 'r') as mf:
            metadata = json.load(mf)

        imageSize = get_image_size(metadata, imagesdir)

        filename = anns['filename'].split('.')[0]
        labelfile = open('{}/{}.txt'.format(outdir, filename), 'w+')
        # for each object in the metadata file, check if any of the properties are true
        for obj in metadata['objects']:
            for prop in mapping['properties']:
                if eval(prop):
                    for ann in anns['annotations']:
                        if ann['id'] == obj['id']:
                            objann = ann
                            break
                    else:  # All the objects from the scene are recorded in metadata; only those in the image are annotated
                        continue

                    cat = mapping['classes'][mapping['properties'][prop]]

                    # Metadata needed to make Kitti labels from annotations:
                    if 'channel' in metadata and metadata['channel'] == 'satrgb':
                        metadata['sensor'].update({
                            'lens_type': 'orthographic',
                            'lens_orthographic_scale': 1.0
                        })
                    if 'sensor' not in metadata:
                        metadata['sensor'] = {
                            'lens_type': 'perspective',
                            'focal_length': 50,
                            'camera_size': 36  # width
                        }

                    label = kitti_label(cat[-1], objann, imageSize, metadata)
                    labelfile.write(label + '\n')

                    break

        labelfile.close()


# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import bpy, os, bpy_extras, mathutils, numpy, cv2, json, logging, imageio
import anatools.lib.context as ctx
from  anatools.lib.camera_checks import collect_mesh_objects

logger = logging.getLogger(__name__)
loglevel = logging.getLogger().level
MIN_FEATURE_SIZE = 2  # Minimum value of 2 - bboxs with boundingRect h=1, have bbox h=0

def compute_polygons(obj):
    """ Generates the polygon from a mask segmentation and bounding box array. """
    maskfile = obj.mask.replace('#', str(obj.active_scene.frame_current))
    poly,bbox = [],None
    img = imageio.imread(maskfile)
    img = numpy.where(img == obj.instance, 255, 0).astype(numpy.uint8)
    contours, _ = cv2.findContours(img, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)
    allpts = []
    for c in contours:
        a = c.flatten()
        pts = []
        for i in a: 
            pts.append(int(i))
            allpts.append(i)
        if len(pts) > 4: poly.append(pts)
    polyarr = numpy.array(allpts, dtype=numpy.int32)
    polyarr = polyarr.reshape(int(len(polyarr)/2), 2)
    if len(polyarr) < 3: return None, None
    x,y,w,h = cv2.boundingRect(polyarr)
    if w < MIN_FEATURE_SIZE or h < MIN_FEATURE_SIZE: poly,bbox = None,None
    else: 
        bbox = [int(val) for val in [x,y,w-1,h-1]]
    return poly, bbox

def total_bound_box(obj):
    #returns a bound box for object and all it's children
    mesh_objects = collect_mesh_objects(obj)
    bbox = mesh_objects[0].bound_box

    xmin = bbox[0][0]
    xmax = bbox[0][0]
    ymin = bbox[0][1]
    ymax = bbox[0][1]
    zmin = bbox[0][2]
    zmax = bbox[0][2]

    for mesh_object in mesh_objects:

        for coord in mesh_object.bound_box:
            if coord[0] < xmin: xmin=coord[0]
            if coord[0] > xmax: xmax=coord[0]
            if coord[1] < ymin: ymin=coord[1]
            if coord[1] > ymax: ymax=coord[1]
            if coord[2] < zmin: zmin=coord[2]
            if coord[2] > zmax: zmax=coord[2]

    return [xmin,xmax,ymin,ymax,zmin,zmax]




def compute_bbox3d(obj):
    blendobj = obj.root
    if blendobj.type in ['EMPTY', 'ARMATURE']: blendobj = blendobj.children[0]
    bboxcoords = [blendobj.matrix_world @ mathutils.Vector(coord) for coord in blendobj.bound_box]
    bbox3d = []
    for coord in bboxcoords:
        camcoord = bpy_extras.object_utils.world_to_camera_view(obj.active_scene, obj.active_scene.camera, coord)
        bbox3d.append( int(camcoord[0]*obj.active_scene.render.resolution_x) )
        bbox3d.append( int((1-camcoord[1])*obj.active_scene.render.resolution_y) )
        bbox3d.append( camcoord[2] )
    return [val for val in bbox3d]


def compute_centroid(obj):
    blendobj = obj.root
    if blendobj.type in ['EMPTY', 'ARMATURE']: blendobj = blendobj.children[0]
    bboxcoords = [blendobj.matrix_world @ mathutils.Vector(coord) for coord in blendobj.bound_box]
    center = sum((mathutils.Vector(b) for b in bboxcoords), mathutils.Vector()) / 8
    camcoord = bpy_extras.object_utils.world_to_camera_view(obj.active_scene, obj.active_scene.camera, center)
    x = int(camcoord[0]*obj.active_scene.render.resolution_x)
    y = int((1-camcoord[1])*obj.active_scene.render.resolution_y)
    return [y,x], camcoord[2]


def compute_rle(obj):
    maskfile = obj.mask.replace('#', str(obj.active_scene.frame_current))
    img = imageio.imread(maskfile)
    rle = { 'size':img.shape, 'counts':[] }
    count = 1
    prev = [0,0,0]
    for y in range(img.shape[0]):
        for x in range(img.shape[1]):
            if not (img[y,x] == prev).all():
                rle['counts'].append(count)
                prev = img[y,x]
                count = 1
            else: count += 1
    return rle


def truncated(obj, bbox):
    if bbox is None:
        return
    x, y, w, h = bbox
    if x == 0 or y == 0:
        return True
    maskfile = obj.mask.replace('#', str(obj.active_scene.frame_current))
    img = imageio.imread(maskfile)
    if x+w+1 == img.shape[0] or y+h+1 == img.shape[1]:
        return True
    return False


def compute_size(obj):
    """ Return x,y,z or length, width, depth. """
    return [c for c in obj.root.dimensions]


# https://automaticaddison.com/how-to-convert-a-quaternion-into-euler-angles-in-python/
def euler_from_quaternion(x, y, z, w):
    """
    Convert a quaternion into euler angles (roll, pitch, yaw)
    roll is rotation around x in radians (counterclockwise)
    pitch is rotation around y in radians (counterclockwise)
    yaw is rotation around z in radians (counterclockwise)
    """
    t0 = +2.0 * (w * x + y * z)
    t1 = +1.0 - 2.0 * (x * x + y * y)
    roll_x = numpy.arctan2(t0, t1)

    t2 = +2.0 * (w * y - z * x)
    t2 = +1.0 if t2 > +1.0 else t2
    t2 = -1.0 if t2 < -1.0 else t2
    pitch_y = numpy.arcsin(t2)

    t3 = +2.0 * (w * z + x * y)
    t4 = +1.0 - 2.0 * (y * y + z * z)
    yaw_z = numpy.arctan2(t3, t4)

    return [roll_x, pitch_y, yaw_z]  # in radians


def compute_rotation(obj):
    """ Return Euler angles: roll, pitch, and yaw. """
    loc, rot, scale = obj.root.matrix_world.decompose()

    roll, pitch, yaw = None, None, None
    # Test me
    # if type(rot) == mathutils.Euler:
    #     rot.order = 'XYX'
    #     roll, pitch, yaw = rot

    if type(rot) == mathutils.Quaternion:
        w, x, y, z = rot
        roll, pitch, yaw = euler_from_quaternion(x, y, z, w)

    return roll, pitch, yaw


def compute_obstruction(obj):
    """Estimate the amount of an object's mask is hidden from view of the camera. 0 - full view; 1 - out of view. """
    maskfilebase, maskext = obj.mask.replace('#', str(obj.active_scene.frame_current)).rsplit('.', 1)
    solomaskfile = '{}-{}.{}'.format(maskfilebase, obj.solo_mask_id, maskext)

    if obj.solo_mask_id:
        soloimg = imageio.imread(solomaskfile)
    else:
        logger.warning('Object {} has no mask'.format(obj.instance))
        return

    compositemaskfile = obj.mask.replace('#', str(obj.active_scene.frame_current))
    compimg = imageio.imread(compositemaskfile)
    compmask = numpy.nonzero(compimg == obj.instance)[0]

    solononzero = soloimg[numpy.nonzero(soloimg)]
    pix_ids = numpy.unique(solononzero)  # the second entry is the first entry + obj.instance
    if len(pix_ids)<2:
        logger.warning('Object {} not in solo mask but has {} px in the original image'.format(obj.instance, str(len(compmask))))
        return
    
    solomask = numpy.nonzero(soloimg == pix_ids[1])[0]
    obstruction = 1.0 - len(compmask) / len(solomask)

    return obstruction



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import logging
import bpy

logger = logging.getLogger(__name__)


def get_child_objects(model):
    """Return a list of all children of an object"""

    children = []
    for child in model.children:
        children.append(child)
    return children


def find_object(model, qname):
    """
    Search an object hierarchy for a qualified object name. Return a pointer to the object. 
    Example:

        find_object(
            bpy.data.objects["Boeing_E3.003"],
            ["Boeing_E3", "rotodome"]
        )

        will return the "rotodome.003" object in the "Boeing_E3.003" hierarchy.

    """
    
    if not model.name.startswith(qname[0]):
        # top object in model doesn't match qname
        logger.critical("Top object in model '%s' is not named %s", model.name, qname[0])
        sys.exit(1)
    else:
        obj = model
        # find subsequent qnames
        for level in range(1, len(qname)):
            children = get_child_objects(obj)
            found_child = False
            for child in children:
                if child.name.startswith(qname[level]):
                    found_child = True
                    obj = child
                    break
            if not found_child:
                # no child matched qname at this level
                logger.critical("Model '%s' does not contain an object named '%s'", model.name, ".".join(qname[:level+1]))
                sys.exit(1)
    # if all levels matched then we found the object
    return obj

def find_mesh(model, qname, mesh_prefix):
    """Find a mesh that is attached to a qualified object name"""

    # find the object by its qualified name
    obj = find_object(model, qname)

    # check if any meshes matching the prefix are associated with the target object
    mesh_found = None
    for mesh in bpy.data.meshes:
        if mesh.name.startswith(mesh_prefix):
            if obj.data == mesh:
                mesh_found = mesh

    # either no mesh matched the prefix or none of the matching meshes were associated with the object
    if not mesh_found:
        logger.critical("Object '%s' does not have mesh '%s'", " ".join(qname), mesh_prefix)
        sys.exit(1)

    return mesh_found

def find_material(material_name):
    """Find material by name"""

    for mat in bpy.data.materials:
        if mat.name == material_name:
            return mat

    # couldn't find a material with that name
    logger.critical("Couldn't find material named '%s'", material_name)
    sys.exit(1)

def find_root(collection):
    """Find the root object in a collection"""

    # find the object that has no parent
    children = []
    all_objects = list(collection.all_objects)
    for parent in all_objects:
        children.extend(list(parent.children))
    roots = list(set(all_objects) - set(children))

    # there should only be one root
    if len(roots) == 0:
        logger.critical("No root found in collection '%s'", collection.name)
        sys.exit(1)
    elif len(roots) > 1:
        logger.critical("Multiple roots found in collection '%s'", collection.name)
        sys.exit(1)

    return roots[0]



"""
Helper functions to parse API calls.
"""

def errorhandler(self, response, call):
    responsedata = response.json()
    ret = False
    api_failure_message = f'API {call} failed.'
    
    if self.verbose == 'debug': print(responsedata)
    
    if 'data' not in responsedata: 
        if self.verbose: print(api_failure_message)
    elif responsedata['data'] is None:
        if 'errors'in responsedata and 'message' in responsedata['errors'][0]:
            print(responsedata['errors'][0]['message'])
        else:
            if self.verbose: print(api_failure_message)
    elif call not in responsedata['data']:
        if self.verbose: print(api_failure_message)
    elif responsedata['data'][call] is None:
        if 'errors'in responsedata and 'message' in responsedata['errors'][0]:
            print(responsedata['errors'][0]['message'])
        else:
            if self.verbose: print(api_failure_message)
    else:
        if responsedata['data'][call] == "success": ret = True
        elif responsedata['data'][call] == "failure": ret = api_failure_message
        else: ret = responsedata['data'][call]
        
    return ret



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from anatools.lib.node import Node

logger = logging.getLogger(__name__)


class Value(Node):

    def exec(self):
        logger.info("Executing {}".format(self.name))
        # parse inputs
        value = float(self.inputs['Value'][0])
        return {"Value": value}


class String(Node):

    def exec(self):
        logger.info("Executing {}".format(self.name))
        # parse inputs
        string = str(self.inputs['String'][0])
        return {"String": string}


"""
Analytics API calls.
"""

def getAnalytics(self, analyticsId, datasetId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getAnalytics",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "analyticsId": analyticsId
            },
            "query": """query 
                getAnalytics($workspaceId: String!, $datasetId: String!, $analyticsId: String!) {
                    getAnalytics(workspaceId: $workspaceId, datasetId: $datasetId, analyticsId: $analyticsId){
                        analyticsId
                        workspaceId
                        datasetId
                        status
                        result
                        member
                        type
                        range
                        returnImages
                    }
                }"""})
    return self.errorhandler(response, "getAnalytics")


def getAnalyticsTypes(self):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getAnalyticsTypes",
            "variables": {},
            "query": """query 
                getAnalyticsTypes{
                    getAnalyticsTypes
                }"""})
    return self.errorhandler(response, "getAnalyticsTypes")


def createAnalytics(self, workspaceId, datasetId, rangeInput, typeInput, returnImages):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createAnalytics",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "range": rangeInput,
                "type": typeInput,
                "returnImages": returnImages
            },
            "query": """mutation 
                createAnalytics($workspaceId: String!, $datasetId: String!, $range: [Int], $type: String!, $returnImages: Boolean) {
                    createAnalytics(workspaceId: $workspaceId, datasetId: $datasetId, range: $range, type: $type, returnImages: $returnImages)
                }"""})
    return self.errorhandler(response, "createAnalytics")


def deleteAnalytics(self, workspaceId, analyticsId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteAnalytics",
            "variables": {
                "workspaceId": workspaceId,
                "analyticsId": analyticsId
            },
            "query": """mutation 
                deleteAnalytics($workspaceId: String!, $datasetId: String! $analyticsId: String!) {
                    deleteAnalytics(workspaceId: $workspaceId, datasetId: $datasetId, analyticsId: $analyticsId)
                }"""})
    return self.errorhandler(response, "deleteAnalytics")



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json
import logging
import datetime
import bpy
import cv2
import numpy
import anatools.lib.context as ctx

logger = logging.getLogger(__name__)
loglevel = logging.getLogger().level

# pylint: disable=method-hidden
# pylint: disable=W0221

class MetadataEncoder(json.JSONEncoder):
    """ Custom encoder to convert metadata to JSON """
    def default(self, obj):
        if hasattr(obj, 'dump_metadata'):
            return obj.dump_metadata()
        return json.JSONEncoder.default(self, obj)


class AnaScene:
    """ Base class for a scene """
    def __init__(self, blender_scene=None, annotation_view_layer=None, objects=None, sensor_name="Image"):
        """ initialize scene """
        self.filename = None # this is set when annotations are written
        self.sensor_name = sensor_name
        self.blender_scene = blender_scene
        if annotation_view_layer is None:
            self.annotation_view_layer = bpy.context.view_layer
        else:
            self.annotation_view_layer = annotation_view_layer
        self.configure_compositor()

        self.objects = []
        if objects is not None:
            for obj in objects:
                self.add_object(obj)

    def dump_metadata(self):
        """ Dump metadata """
        # only show metadata for objects that were rendered
        rendered_objects = []
        for obj in self.objects:
            if obj.rendered:
                rendered_objects.append(obj)
        return {
            "filename": self.filename,
            "channel": ctx.channel.name,
            "version": '0.0.1',
            "date": datetime.datetime.now().isoformat(),
            "objects": rendered_objects}
   
    def configure_compositor(self):
        """ configure the compositor """
        self.blender_scene.use_nodes = True
        self.annotation_view_layer.use_pass_object_index = True
        nodes = self.blender_scene.node_tree.nodes
        links = self.blender_scene.node_tree.links
        filename = f'{ctx.interp_num:010}-#-{self.sensor_name}'

        self.imgout = nodes.new('CompositorNodeOutputFile') #output image node
        self.imgout.name = 'imgout'
        self.imgout.base_path = os.path.join(ctx.output, "images")
        self.imgout.format.file_format = "PNG"
        self.imgout.format.color_depth = "8"
        self.imgout.format.compression = 0
        self.imgout.file_slots.clear()
        self.imgout.file_slots.new(filename)

        self.maskout = nodes.new('CompositorNodeOutputFile') #output mask node
        self.maskout.name = 'maskout'
        self.maskout.base_path = os.path.join(ctx.output, "masks")
        self.maskout.format.file_format = "PNG"
        self.maskout.format.color_mode = "BW"
        self.maskout.format.color_depth = "16"
        self.maskout.format.compression = 0
        self.maskout.file_slots.remove(self.maskout.inputs['Image'])
        self.maskout.file_slots.new(filename)
        self.maskoutput = None
        self.mask = os.path.join(ctx.output, 'masks', filename)

        self.last_output = nodes['Render Layers'].outputs['Image']
        self.last_link = links.new(self.last_output, self.imgout.inputs[0])

    def add_object(self, obj, ooi=True):
        """ add an object to the scene """
        self.objects.append(obj)
        self.configure_mask(obj)
        obj.rendered = True
        obj.ooi = ooi  #Default is that the objects added to the Anascene are objects of interest.
        obj.active_scene = self.blender_scene

    def configure_mask(self, obj):
        """ set the pass index for the object and each of its subobjects """
        if not self.blender_scene.use_nodes:
            logger.info("AnaScene.configure_mask: Something's wrong!")
            self.configure_compositor()

        nodes = self.blender_scene.node_tree.nodes
        links = self.blender_scene.node_tree.links

        obj.root.pass_index = obj.instance
        children = [child.name for child in obj.root.children]
        while len(children):
            bpy.data.objects[children[0]].pass_index = obj.instance
            for child in bpy.data.objects[children[0]].children:
                children.append(child.name)
            children.pop(0)

        # maskroot = f'{ctx.interp_num:010}-{obj.root.name}-#'
        # self.maskout.file_slots.new(maskroot)

        logger.debug(f"Adding IDMask Node for {obj.root.name}")
        masknode = nodes.new('CompositorNodeIDMask')
        masknode.name = obj.root.name + '_mask'
        masknode.index = obj.instance
        links.new(nodes['Render Layers'].outputs['IndexOB'], masknode.inputs['ID value'])
        dividenode = nodes.new('CompositorNodeMath')
        dividenode.name = obj.root.name +'_divide'
        dividenode.operation = 'DIVIDE'
        dividenode.inputs[1].default_value = 65535/obj.instance
        links.new(masknode.outputs['Alpha'], dividenode.inputs[0])
        if self.maskoutput is None:
            self.maskoutput = dividenode.outputs[0]
            links.new(dividenode.outputs[0], self.maskout.inputs[0])
        else:
            links.remove(self.maskout.inputs[0].links[0])
            maxnode = nodes.new('CompositorNodeMath')
            maxnode.name = obj.root.name + '_max'
            maxnode.operation = 'MAXIMUM'
            links.new(self.maskoutput, maxnode.inputs[0])
            links.new(dividenode.outputs[0], maxnode.inputs[1])
            links.new(maxnode.outputs[0], self.maskout.inputs[0])
            self.maskoutput = maxnode.outputs[0]

        if not os.path.isdir(os.path.join(ctx.output, 'masks')):
            os.mkdir(os.path.join(ctx.output, 'masks'))


    def write_ana_annotations(self, calculate_obstruction=False):
        """ Creates an annotations file of the image in <output>/annotations/{imgfile}-anatools.json """
        if not self.filename: self.filename = f'{ctx.interp_num:010}-{self.blender_scene.frame_current}-{self.sensor_name}.png'
        if not os.path.isdir(os.path.join(ctx.output, 'annotations')):
            os.mkdir(os.path.join(ctx.output, 'annotations'))
        annfile = os.path.join(ctx.output, 'annotations', f'{ctx.interp_num:010}-{self.blender_scene.frame_current}-{self.sensor_name}-ana.json')

        logger.info(f"Writing annotations to: {annfile}")

        # generate list of annotations
        ann_list = []
        for obj in self.objects:
            if obj.rendered and obj.ooi:
                logger.debug(f"Generated annotation for {obj.root.name}")
                obj.mask = os.path.join(ctx.output, f'masks/{ctx.interp_num:010}-#-{self.sensor_name}.png')
                ann = obj.dump_annotations(calculate_obstruction=calculate_obstruction)
                if ann: ann_list.append(ann)

        annotation_out = {
            "filename": self.filename,
            "annotations": ann_list
        }

        with open(annfile, 'w') as f:
            json.dump(annotation_out, f, indent=4)


    def write_ana_metadata(self):
        """ Creates a metadata file of the image in <output>/metadata/{filename}-meta.json """
        if not os.path.isdir(os.path.join(ctx.output, 'metadata')):
            os.mkdir(os.path.join(ctx.output, 'metadata'))
        metafile = os.path.join(ctx.output, 'metadata', f'{ctx.interp_num:010}-{self.blender_scene.frame_current}-{self.sensor_name}-metadata.json')

        logger.info(f"Writing metadata to: {metafile}")

        with open(metafile, "w") as f:
            json.dump(self, f, cls=MetadataEncoder, indent=4)



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
import json
import os
import copy
import numpy as np
import anatools.lib.context as ctx
from anatools.lib.package_utils import get_volume_path
from anatools.lib.ana_object import AnaObject

def get_blendfile_generator(package, object_class, object_type):
    """
    Helper function that creates a generator from an object
    definition in the package.yml file
    """
    rel_path = ctx.packages[package]["objects"][object_type]["filename"]
    blender_file = get_volume_path(package, rel_path)
    generator = ObjectGenerator(
        object_class,
        object_type,
        blender_file=blender_file,
        config=ctx.packages[package]["objects"][object_type])
    return generator

class Generator(ABC):
    """
    Base class for generators and modifiers
    """
    next_id = 0
    # all generators indexed by id
    generators = {}
    def __init__(self, children=None, **kwargs):
        self.id = Generator.next_id
        Generator.generators[self.id] = self
        Generator.next_id += 1
        if children is None:
            self.children = []
        else:
            self.children = children
        self.kwargs = kwargs
        self.weight = 1

    @abstractmethod
    def exec(self, *args, **kwargs):
        pass

    def clone(self):
        """
        Clone a generator. The clone has no children.
        """
        the_clone = copy.copy(self)

        the_clone.id = Generator.next_id
        Generator.generators[the_clone.id] = the_clone
        Generator.next_id += 1
        the_clone.children = []
        return the_clone

    def select_child(self):
        """ Select a weighted random child """
        weights = []
        for child in self.children:
            weights.append(child.weight)
        weights = np.array(weights)
        selected = ctx.random.choice(self.children, p=weights/sum(weights))
        return selected

class ObjectGenerator(Generator):
    """
    Object Generator
    """
    def __init__(self, object_class, object_type, **kwargs):
        """ Note: kwargs are for the loader """
        super().__init__(children=[], **kwargs)
        self.object_class = object_class
        self.object_type = object_type

    def exec(self, *args, **kwargs):
        """ Return a new instance of the specified object """
        # instantiate the object class
        obj = self.object_class(self.object_type)
        # load the object into the scene
        obj.load(**self.kwargs)
        # return the object
        return obj

    def __repr__(self):
        return json.dumps({
            "class": self.__class__.__name__,
            "id": self.id,
            "object_class": self.object_class.__name__,
            "object_type": self.object_type,
            "kwargs": self.kwargs
        })

class ObjectModifier(Generator):
    """
    Object Modifier
    """
    def __init__(self, method, children, **kwargs):
        """ Note: kwargs are for the modifier method """
        super().__init__(children=children, **kwargs)
        self.method = method

    def exec(self, *args, **kwargs):
        """ Execute modifier method on object """
        child = self.select_child()
        # recursively execute until we get an object
        if not isinstance(child, AnaObject):
            child = child.exec(*args, **kwargs)
        # execute modifier method
        getattr(child, self.method)(**self.kwargs)
        return child

    def __repr__(self):
        return json.dumps({
            "class": self.__class__.__name__,
            "id": self.id,
            "method": self.method,
            "kwargs": self.kwargs
        })

class CreateBranchGenerator(Generator):
    """
    A generator that connects branches together
    """
    def __init__(self, children, **kwargs):
        super().__init__(children=children, **kwargs)

    def exec(self, *args, **kwargs):
        """ Select a child and execute it """
        child = self.select_child()
        result = child.exec(*args, **kwargs)
        return result

    def __repr__(self):
        return json.dumps({
            "class": self.__class__.__name__,
            "id": self.id
        })

class PathList(list):
    """
    A list of paths. Each path is a list of generator id's from root to leaf
    """
    def __init__(self, paths=None):
        if paths is None:
            super().__init__([])
        else:
            super().__init__(paths)

    def to_tree(self):
        """
        Convert a PathList into an executable tree
        """
        # clone the generators
        new_generators = {}
        id_map = {}
        new_children = {}
        for path in self:
            for gen_id in path:
                if gen_id not in new_generators:
                    clone = Generator.generators[gen_id].clone()
                    new_generators[gen_id] = clone
                    id_map[gen_id] = clone.id
                    new_children[clone.id] = []
        # create a map of the children
        for path in self:
            if len(path) > 1:
                for i in range(len(path[:-1])):
                    clone_id = id_map[path[i]]
                    clone_child_id = id_map[path[i+1]]
                    if clone_child_id not in new_children[clone_id]:
                        new_children[clone_id].append(Generator.generators[clone_child_id])
        # update the children of the clones
        for key, value in new_children.items():
            Generator.generators[key].children = value

        tree = Generator.generators[id_map[self[0][0]]]
        return tree

def _get_all_leaves(tree, leaf_class=ObjectGenerator, leaves=None):
    """
    Helper function that gets a list of all leaves in the tree. The list may include
    dupes because it follows every possible path
    """
    if leaves is None:
        leaves = []
    if isinstance(tree, leaf_class):
        leaves.append(tree)
        return leaves
    else:
        for child in tree.children:
            leaves = _get_all_leaves(child, leaf_class, leaves)
        return leaves

def get_unique_leaves(tree, leaf_class=ObjectGenerator):
    """
    Get list of all unique leaves in the tree.
    """
    leaves_with_dupes = _get_all_leaves(tree, leaf_class=leaf_class)
    ids = []
    leaves = []
    for leaf in leaves_with_dupes:
        if leaf.id not in ids:
            ids.append(leaf.id)
            leaves.append(leaf)
    return leaves

def _get_single_pathlist(tree, leaf_class=ObjectGenerator):
    """
    Helper function that follows a random path to a leaf.
    Returns a PathList with a single path
    """
    path = PathList([[]])
    while not isinstance(tree, leaf_class):
        path[0].append(tree.id)
        tree = tree.select_child()
    path[0].append(tree.id)
    return path

def create_single_path(tree, leaf_class=ObjectGenerator):
    """
    Create an exectuable single path to a weighted random leaf.
    """
    return _get_single_pathlist(tree, leaf_class=ObjectGenerator).to_tree()

def _get_all_paths(tree, leaf_class=ObjectGenerator, paths=None, current_depth=0):
    """
    Helper function that creates a list of all paths in the tree
    Returns a Pathlist with one entry per path
    """
    if paths is None:
        paths = PathList([[]])
    current_depth += 1
    paths[-1].append(tree.id)
    if isinstance(tree, leaf_class):
        # if we are at a leaf then we're done with this path
        return paths
    else:
        # otherwise continue down the left-most path until we hit a leaf
        paths = _get_all_paths(tree.children[0], leaf_class, paths, current_depth)
        if len(tree.children) > 1:
            # if there are more children then create new path
            for child in tree.children[1:]:
                paths.append(paths[-1][:current_depth])
                paths = _get_all_paths(child, leaf_class, paths, current_depth)
        return paths

def create_multi_path(tree, leaf_class=ObjectGenerator):
    """
    Create an executable multi path to a weighted random leaf. Path includes all
    possible routes to the leaf.
    """
    # generate a pathlist
    single_path = _get_single_pathlist(tree, leaf_class)
    # get the leaf
    leaf_id = single_path[0][-1]
    # generate all possible paths
    all_paths = _get_all_paths(tree, leaf_class)
    # save all paths that end at the leaf
    paths_to_leaf = PathList()
    for path in all_paths:
        if path[-1] == leaf_id:
            paths_to_leaf.append(path)
    # convert back to a tree
    return paths_to_leaf.to_tree()



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" 
A collection of utility functions that load assets from existing blender files 
"""
import bpy
from anatools.lib.search_utils import find_root

def load_model(blender_file_name, collection_name):
    '''
    Load a model. Assumes the model is in a collection and there
    is a single parent object for the model.
    Returns a pointer to the parent object.
    '''

    # load the collection
    new_collection = load_collection(blender_file_name, collection_name)
    
    # link it to the current scene
    bpy.context.scene.collection.children.link(new_collection)

    # find the root of the collection - there should only be one
    root = find_root(new_collection)

    return root

def load_collection(blender_file_name, collection_name):
    '''
    Load a collection.
    Returns a pointer to the collection.
    '''

    # load collection
    with bpy.data.libraries.load(filepath="//" + blender_file_name, link=False) as (_, dt):
        dt.collections = [collection_name]

    return dt.collections[0]

def load_material(blender_file_name, material_name):
    """
    Load a material from a blender file.
    Returns a pointer to the material.
    """

    with bpy.data.libraries.load(filepath="//" + blender_file_name, link=False) as (_, dt):
        dt.materials = [material_name]
    return dt.materials[0]


def load_text(blender_file_name, text_name):
    """
    Load a text from a blender file.
    Returns a pointer to the text file.
    """

    with bpy.data.libraries.load(filepath="//" + blender_file_name, link=False) as (_, dt):
        dt.texts = [text_name]
    return dt.texts[0]